@@ -134,7 +134,11 @@ typedef enum {
     /* Note: if you're hunting for a flag, some day in the future when we
      * have used them all, this one is easy enough to eliminate by having the
      * tiny number of objects marked this way in a remembered set. */
-    MVM_CF_NEVER_REPOSSESS = 2048
+    MVM_CF_NEVER_REPOSSESS = 2048,
+
+    /* Has this item been chained into a gen2 freelist? This is only used in
+     * GC debug more. */
+    MVM_CF_DEBUG_IN_GEN2_FREE_LIST = 4096,
 } MVMCollectableFlags;
 
 #ifdef MVM_USE_OVERFLOW_SERIALIZATION_INDEX
@@ -473,6 +477,13 @@ struct MVMREPROps_Positional {
     void (*shift) (MVMThreadContext *tc, MVMSTable *st,
         MVMObject *root, void *data, MVMRegister *value, MVMuint16 kind);
 
+    /* Creates a slice of the source array (from start to end) and stores it
+     * within the destination array. If start or end is "-n", it will be
+     * translated into the nth position relative to the end of the array. */
+    void (*slice) (MVMThreadContext *tc, MVMSTable *st,
+        MVMObject *src, void *data, MVMObject *dest,
+        MVMint64 start, MVMint64 end);
+
     /* Splices the specified array into this one. Representations may optimize if
      * they know the type of the passed array, otherwise they should use the REPR
      * API. */
