@@ -58,6 +58,14 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CSTRUCT;
             repr_data->elem_size = sizeof(void *);
         }
+        else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+            repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CPPSTRUCT;
+            repr_data->elem_size = sizeof(void *);
+        }
+        else if (type_id == MVM_REPR_ID_MVMCUnion) {
+            repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CUNION;
+            repr_data->elem_size = sizeof(void *);
+        }
         else if (type_id == MVM_REPR_ID_MVMCPointer) {
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CPOINTER;
             repr_data->elem_size = sizeof(void *);
@@ -121,12 +129,10 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
     MVMCArrayBody *body = (MVMCArrayBody *)data;
 
-    if (body->managed) {
+    if (body->managed) 
         MVM_free(body->storage);
-
-        if (body->child_objs)
-            MVM_free(body->child_objs);
-    }
+    if (body->child_objs)
+        MVM_free(body->child_objs);
 }
 
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
@@ -199,6 +205,8 @@ static void expand(MVMThreadContext *tc, MVMCArrayREPRData *repr_data, MVMCArray
     is_complex = (repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CARRAY
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CPOINTER
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CSTRUCT
+               || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CPPSTRUCT
+               || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CUNION
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_STRING);
 
     if (is_complex) {
@@ -373,6 +381,18 @@ static void bind_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
             bind_wrapper_and_ptr(tc, root, body, index, value.o,
                 IS_CONCRETE(value.o) ? ((MVMCStruct *)value.o)->body.cstruct : NULL);
             break;
+        case MVM_CARRAY_ELEM_KIND_CPPSTRUCT:
+            if (REPR(value.o)->ID != MVM_REPR_ID_MVMCPPStruct)
+                MVM_exception_throw_adhoc(tc, "CArray of CPPStruct passed non-CStruct object");
+            bind_wrapper_and_ptr(tc, root, body, index, value.o,
+                IS_CONCRETE(value.o) ? ((MVMCPPStruct *)value.o)->body.cppstruct : NULL);
+            break;
+        case MVM_CARRAY_ELEM_KIND_CUNION:
+            if (REPR(value.o)->ID != MVM_REPR_ID_MVMCUnion)
+                MVM_exception_throw_adhoc(tc, "CArray of CUnion passed non-CStruct object");
+            bind_wrapper_and_ptr(tc, root, body, index, value.o,
+                IS_CONCRETE(value.o) ? ((MVMCUnion *)value.o)->body.cunion : NULL);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown element type in CArray");
     }
@@ -394,6 +414,10 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     die_pos_nyi(tc);
 }
 
+static void aslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *data, MVMObject *dest, MVMint64 start, MVMint64 end) {
+    die_pos_nyi(tc);
+}
+
 static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMCArrayBody *body = (MVMCArrayBody *)data;
 
@@ -457,12 +481,15 @@ static const MVMREPROps CArray_this_repr = {
         pop,
         unshift,
         shift,
+        aslice,
         MVM_REPR_DEFAULT_SPLICE,
         MVM_REPR_DEFAULT_AT_POS_MULTIDIM,
         MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
         MVM_REPR_DEFAULT_DIMENSIONS,
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
-        MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC
+        MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC,
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     elems,
