@@ -109,6 +109,9 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
     else if (tc->instance->debugserver)
         MVM_breakpoint_instrument(tc, static_frame);
     else
+        /* XXX uninstrumenting is currently turned off, due to multithreading
+         * woes. If you add an instrumentation that has to be "turned off"
+         * again at some point, a solution for this problem must be found. */
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
@@ -158,31 +161,31 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
         frame->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, static_frame->body.env_size);
         frame->allocd_env = static_frame->body.env_size;
         if (autoclose) {
-            MVMuint16 i, num_lexicals = static_frame->body.num_lexicals;
+            MVMROOT2(tc, frame, static_frame, {
+                MVMuint16 i;
+                MVMuint16 num_lexicals = static_frame->body.num_lexicals;
+                for (i = 0; i < num_lexicals; i++) {
+                    if (!static_frame->body.static_env[i].o && static_frame->body.static_env_flags[i] == 1) {
+                        MVMint32 scid;
+                        MVMint32 objid;
+                        if (MVM_bytecode_find_static_lexical_scref(tc, static_frame->body.cu,
+                                static_frame, i, &scid, &objid)) {
+                            MVMObject *resolved;
+                            MVMSerializationContext *sc = MVM_sc_get_sc(tc, static_frame->body.cu, scid);
 
-            MVM_gc_root_temp_push(tc, (MVMCollectable **)&static_frame);
+                            if (sc == NULL)
+                                MVM_exception_throw_adhoc(tc,
+                                    "SC not yet resolved; lookup failed");
 
-            for (i = 0; i < num_lexicals; i++) {
-                if (!static_frame->body.static_env[i].o && static_frame->body.static_env_flags[i] == 1) {
-                    MVMint32 scid, objid;
-                    if (MVM_bytecode_find_static_lexical_scref(tc, static_frame->body.cu,
-                            static_frame, i, &scid, &objid)) {
-                        MVMObject *resolved;
-                        MVMSerializationContext *sc = MVM_sc_get_sc(tc, static_frame->body.cu, scid);
+                            resolved = MVM_sc_get_object(tc, sc, objid);
 
-                        if (sc == NULL)
-                            MVM_exception_throw_adhoc(tc,
-                                "SC not yet resolved; lookup failed");
-
-                        resolved = MVM_sc_get_object(tc, sc, objid);
-
-                        MVM_ASSIGN_REF(tc, &(static_frame->common.header),
-                            static_frame->body.static_env[i].o,
-                            resolved);
+                            MVM_ASSIGN_REF(tc, &(static_frame->common.header),
+                                static_frame->body.static_env[i].o,
+                                resolved);
+                        }
                     }
                 }
-            }
-            MVM_gc_root_temp_pop(tc);
+            });
         }
         memcpy(frame->env, static_frame->body.static_env, static_frame->body.env_size);
     }
@@ -1208,8 +1211,8 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
             MVMROOT(tc, f, {
                 resolved = MVM_sc_get_object(tc, sc, objid);
             });
-            MVM_ASSIGN_REF(tc, &(f->static_info->common.header),
-                f->static_info->body.static_env[effective_idx].o,
+            MVM_ASSIGN_REF(tc, &(effective_sf->common.header),
+                effective_sf->body.static_env[effective_idx].o,
                 resolved);
         }
     }
