@@ -229,12 +229,13 @@ struct MVMInstance {
      * the item is unresolved. Also, array of all SCs, used for the
      * index stored in object headers. When an SC goes away this is
      * simply nulled. That makes it a small memory leak if a lot of
-     * SCs are created and go away over time. */
+     * SCs are created and go away over time. The mutex protects all
+     * the weakhash and all SCs list. */
     MVMSerializationContextBody  *sc_weakhash;
-    uv_mutex_t                    mutex_sc_weakhash;
     MVMSerializationContextBody **all_scs;
     MVMuint32                     all_scs_next_idx;
     MVMuint32                     all_scs_alloc;
+    uv_mutex_t                    mutex_sc_registry;
 
     /* Mutex to serialize additions of type parameterizations. Global rather
      * than per STable, as this doesn't happen often. */
@@ -351,6 +352,12 @@ struct MVMInstance {
     /* Cache the environment hash */
     MVMObject      *env_hash;
 
+    /* Cache the signal hash */
+    MVMObject      *sig_hash;
+
+    /* Flags indicating the signals available on the host system */
+    MVMuint64       valid_sigs;
+
     /************************************************************************
      * Caching and interning
      ************************************************************************/
@@ -501,4 +508,8 @@ struct MVMInstance {
 
     /* Flag for if NFA debugging is enabled. */
     MVMint8 nfa_debug_enabled;
+
+    /* Hash Secret which is used as the hash seed. This is to avoid denial of
+     * service type attacks. */
+    MVMuint32 hashSecret;
 };
