@@ -15,8 +15,9 @@
 
 /* stuff for uthash */
 #define uthash_fatal(msg) MVM_exception_throw_adhoc(tc, "internal hash error: " msg)
+typedef uint32_t MVMhashv;
 
-#include <uthash.h>
+#include "strings/uthash.h"
 
 /* libuv
  * must precede atomic_ops.h so we get the ordering of Winapi headers right
@@ -141,6 +142,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "gc/finalize.h"
 #include "core/regionalloc.h"
 #include "spesh/dump.h"
+#include "spesh/debug.h"
 #include "spesh/graph.h"
 #include "spesh/codegen.h"
 #include "spesh/candidate.h"
@@ -174,6 +176,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "strings/unicode.h"
 #include "strings/latin1.h"
 #include "strings/windows1252.h"
+#include "strings/shiftjis.h"
 #include "io/io.h"
 #include "io/eventloop.h"
 #include "io/syncfile.h"
